@@ -7,6 +7,7 @@ import base64
 import copy
 import functools
 import ipaddress
+import requests
 import json
 import os
 import os.path
@@ -18,13 +19,23 @@ import uuid
 import pkg_resources
 import pkgutil
 import yaml
+from yaml import SafeLoader
 
 from OpenSSL import crypto
 from .apic_provision import Apic, ApicKubeConfig
 from jinja2 import Environment, PackageLoader
 from os.path import exists
 
-DEFAULT_FLAVOR = "kubernetes-1.10"
+
+# This black magic forces pyyaml to load YAML strings as unicode rather
+# than byte strings in Python 2, thus ensuring that the type of strings
+# is consistent across versions.  From
+# https://stackoverflow.com/a/2967461/3857947.
+def construct_yaml_str(self, node):
+    return self.construct_scalar(node)
+
+
+SafeLoader.add_constructor(u'tag:yaml.org,2002:str', construct_yaml_str)
 
 VERSION_FIELDS = [
     "cnideploy_version",
@@ -34,175 +45,27 @@ VERSION_FIELDS = [
     "openvswitch_version",
 ]
 
-VERSIONS = {
-    "1.6": {
-        "cnideploy_version": "1.6r15",
-        "aci_containers_host_version": "1.6r15",
-        "aci_containers_controller_version": "1.6r15",
-        "opflex_agent_version": "1.6r22",
-        "openvswitch_version": "1.6r12",
-    },
-    "1.7": {
-        "cnideploy_version": "1.7r86",
-        "aci_containers_host_version": "1.7r86",
-        "aci_containers_controller_version": "1.7r86",
-        "opflex_agent_version": "1.7r70",
-        "openvswitch_version": "1.7r24",
-    },
-    "1.9": {
-        "cnideploy_version": "1.9r32",
-        "aci_containers_host_version": "1.9r32",
-        "aci_containers_controller_version": "1.9r32",
-        "opflex_agent_version": "1.9r35",
-        "openvswitch_version": "1.7r24",
-    },
-}
 
-# Known Flavor options:
-# - template_generator: Function that generates the output config
-#       file. Default: generate_kube_yaml.
-# - version_fields: List of config options that must be specified for
-#       the specific version of deployment. Default: VERSION_FIELDS.
-# - vip_pool_required: Whether virtual IP pool needs to be specified.
-#       Default: False.
-# - apic: Dict that is used for configuring ApicKubeConfig
-#       Known sub-options:
-#       - use_kubeapi_vlan: Whether kubeapi_vlan should be used. Default: True.
-#       - tenant_generator: Name of the function to generate tenant objects.
-#             Default: kube_tn.
-#       - associate_aep_to_nested_inside_domain: Whether AEP should be attached
-#             to nested_inside domain. Default: False.
-KubeFlavorOptions = {}
+FLAVORS_PATH = os.path.dirname(os.path.realpath(__file__)) + "/flavors.yaml"
+VERSIONS_PATH = os.path.dirname(os.path.realpath(__file__)) + "/versions.yaml"
 
-CfFlavorOptions = {
-    'apic': {
-        'use_kubeapi_vlan': False,
-        'tenant_generator': 'cloudfoundry_tn',
-        'associate_aep_to_nested_inside_domain': True,
-    },
-    'version_fields': [],
-    'vip_pool_required': True,
-}
 
-DEFAULT_FLAVOR_OPTIONS = KubeFlavorOptions
+with open(VERSIONS_PATH, 'r') as stream:
+    try:
+        doc = yaml.safe_load(stream)
+    except yaml.YAMLError as exc:
+        print(exc)
+    VERSIONS = doc['versions']
 
-FLAVORS = {
-    # Upstream Kubernetes
-    "kubernetes-1.10": {
-        "desc": "Kubernetes 1.10",
-        "default_version": "1.9",
-    },
-    "kubernetes-1.9": {
-        "desc": "Kubernetes 1.9",
-        "default_version": "1.9",
-    },
-    "kubernetes-1.8": {
-        "desc": "Kubernetes 1.8",
-        "default_version": "1.7",
-        "config": {
-            "kube_config": {
-                "use_apps_api": "apps/v1beta2",
-                "use_apps_apigroup": "apps",
-            }
-        }
-    },
-    "kubernetes-1.7": {
-        "desc": "Kubernetes 1.7",
-        "default_version": "1.7",
-        "config": {
-            "kube_config": {
-                "use_rbac_api": "rbac.authorization.k8s.io/v1beta1",
-                "use_apps_api": "extensions/v1beta1",
-                "use_apps_apigroup": "extensions",
-            }
-        }
-    },
-    "kubernetes-1.6": {
-        "desc": "Kubernetes 1.6",
-        "default_version": "1.6",
-        "config": {
-            "kube_config": {
-                "use_rbac_api": "rbac.authorization.k8s.io/v1beta1",
-                "use_apps_api": "extensions/v1beta1",
-                "use_apps_apigroup": "extensions",
-                "use_netpol_annotation": True,
-                "use_netpol_apigroup": "extensions",
-            },
-        }
-    },
-    # Red Hat OpenShift Container Platform
-    "openshift-3.9": {
-        "desc": "Red Hat OpenShift Container Platform 3.9",
-        "default_version": "1.9",
-        "config": {
-            "kube_config": {
-                "use_external_service_ip_allocator": True,
-                "use_privileged_containers": True,
-                "use_openshift_security_context_constraints": True,
-                "use_cnideploy_initcontainer": True,
-                "allow_kube_api_default_epg": True,
-                "kubectl": "oc",
-                "system_namespace": "aci-containers-system",
-            },
-            "aci_config": {
-                "vmm_domain": {
-                    "type": "OpenShift",
-                },
-            },
-        },
-    },
-    "openshift-3.6": {
-        "desc": "Red Hat OpenShift Container Platform 3.6",
-        "default_version": "1.6",
-        "config": {
-            "kube_config": {
-                "use_external_service_ip_allocator": True,
-                "use_privileged_containers": True,
-                "use_openshift_security_context_constraints": True,
-                "use_cnideploy_initcontainer": True,
-                "allow_kube_api_default_epg": True,
-                "use_rbac_api": "v1",
-                "use_apps_api": "extensions/v1beta1",
-                "use_apps_apigroup": "extensions",
-                "use_netpol_apigroup": "extensions",
-                "use_netpol_annotation": True,
-                "kubectl": "oc",
-                "system_namespace": "aci-containers-system",
-            },
-            "aci_config": {
-                "vmm_domain": {
-                    "type": "OpenShift",
-                },
-            },
-        },
-    },
-    # Docker Universal Control Plane (UCP)
-    "docker-ucp-3.0": {
-        "desc": "Docker Universal Control Plane (UCP) 3.0",
-        "default_version": "1.7",
-        "config": {
-            "kube_config": {
-                "use_apps_api": "apps/v1beta2",
-                "use_apps_apigroup": "apps",
-                "use_cluster_role": False,
-                "use_cnideploy_initcontainer": True,
-            }
-        }
-    },
-    # CloudFoundry
-    "cloudfoundry-1.0": {
-        "desc": "CloudFoundry cf-deployment 1.x",
-        "default_version": "1.9",
-        "config": {
-            "aci_config": {
-                "vmm_domain": {
-                    "type": "CloudFoundry",
-                },
-            },
-        },
-        "options": CfFlavorOptions,
-    }
-}
+with open(FLAVORS_PATH, 'r') as stream:
+    try:
+        doc = yaml.safe_load(stream)
+    except yaml.YAMLError as exc:
+        print(exc)
+    DEFAULT_FLAVOR = doc['default_flavor']
+    DEFAULT_FLAVOR_OPTIONS = doc['kubeFlavorOptions']
+    CfFlavorOptions = doc['cfFlavorOptions']
+    FLAVORS = doc['flavors']
 
 
 def info(msg):
@@ -286,12 +149,14 @@ def config_default():
         "net_config": {
             "node_subnet": None,
             "pod_subnet": None,
+            "pod_subnet_chunk_size": 32,
             "extern_dynamic": None,
             "extern_static": None,
             "node_svc_subnet": None,
             "kubeapi_vlan": None,
             "service_vlan": None,
             "service_monitor_interval": 0,
+            "interface_mtu": None,
         },
         "kube_config": {
             "controller": "1.1.1.1",
@@ -322,20 +187,11 @@ def config_user(config_file):
     if config_file:
         if config_file == "-":
             info("Loading configuration from \"STDIN\"")
-            config = yaml.load(sys.stdin)
+            config = yaml.safe_load(sys.stdin)
         else:
             info("Loading configuration from \"%s\"" % config_file)
-
-            # This black magic forces pyyaml to load YAML strings as
-            # unicode rather than byte strings in Python 2, thus
-            # ensuring that the type of strings is consistent across
-            # versions.  From
-            # https://stackoverflow.com/a/2967461/3857947.
-            def construct_yaml_str(self, node):
-                return self.construct_scalar(node)
-            yaml.Loader.add_constructor(u'tag:yaml.org,2002:str', construct_yaml_str)
             with open(config_file, 'r') as file:
-                config = yaml.load(file)
+                config = yaml.safe_load(file)
     if config is None:
         config = {}
     return config
@@ -346,21 +202,26 @@ def config_discover(config, prov_apic):
     if prov_apic is not None:
         apic = get_apic(config)
 
+    orig_infra_vlan = config["net_config"].get("infra_vlan")
     ret = {
         "net_config": {
-            "infra_vlan": None,
+            "infra_vlan": orig_infra_vlan,
         }
     }
+
+    infra_vlan = config["discovered"]["infra_vlan"]
     if apic is not None:
         infra_vlan = apic.get_infravlan()
-        ret["net_config"]["infra_vlan"] = infra_vlan
-        orig_infra_vlan = config["net_config"].get("infra_vlan")
+
+    if infra_vlan is not None:
         if orig_infra_vlan is not None and orig_infra_vlan != infra_vlan:
             warn("ACI infra_vlan (%s) is different from input file (%s)" %
                  (infra_vlan, orig_infra_vlan))
-        if orig_infra_vlan is None:
+        if orig_infra_vlan is None or orig_infra_vlan != infra_vlan:
             info("Using infra_vlan from ACI: %s" %
                  (infra_vlan,))
+        ret["net_config"]["infra_vlan"] = infra_vlan
+
     return ret
 
 
@@ -545,6 +406,22 @@ def config_adjust(args, config, prov_apic, no_random):
     return adj_config
 
 
+def is_valid_mtu(xval):
+    if xval is None:
+        # use default configured on this host
+        return True
+
+    xmin = 1280   # for IPv6
+    xmax = 8900   # leave 100 byte header for VxLAN
+    try:
+        x = int(xval)
+        if xmin <= x <= xmax:
+            return True
+    except ValueError:
+        pass
+    raise(Exception("Must be integer between %d and %d" % (xmin, xmax)))
+
+
 def config_validate(flavor_opts, config):
     def Raise(exception):
         raise exception
@@ -590,6 +467,8 @@ def config_validate(flavor_opts, config):
                                      required),
         "net_config/node_svc_subnet": (get(("net_config", "node_svc_subnet")),
                                        required),
+        "net_config/interface_mtu": (get(("net_config", "interface_mtu")),
+                                     is_valid_mtu),
     }
 
     # Allow deletion of resources without isname check
@@ -693,8 +572,12 @@ def config_validate_preexisting(config, prov_apic):
 
 def generate_sample(filep):
     data = pkgutil.get_data('acc_provision', 'templates/provision-config.yaml')
-    filep.write(data)
-    filep.flush()
+    try:
+        filep.write(data)
+    except TypeError:
+        filep.write(data.decode(filep.encoding))
+    finally:
+        filep.flush()
     return filep
 
 
@@ -922,7 +805,7 @@ class CustomFormatter(argparse.HelpFormatter):
         return ret
 
 
-def parse_args():
+def parse_args(show_help):
     version = 'Unknown'
     try:
         version = pkg_resources.require("acc_provision")[0].version
@@ -936,6 +819,8 @@ def parse_args():
     )
     parser.add_argument(
         '-v', '--version', action='version', version=version)
+    parser.add_argument(
+        '--release', action='store_true', default=False, help='print git release info')
     parser.add_argument(
         '--debug', action='store_true', default=False,
         help='enable debug')
@@ -972,9 +857,35 @@ def parse_args():
     parser.add_argument(
         '-t', '--version-token', default=None, metavar='token',
         help='set a configuration version token.  Default is UUID.')
+
+    # If the input has no arguments, show help output and exit
+    if show_help:
+        parser.print_help(sys.stderr)
+        sys.exit(1)
+
     return parser.parse_args()
 
 
+def get_versions(versions_url):
+    global VERSIONS
+    try:
+        # try as a URL
+        res = requests.get(versions_url)
+        versions_yaml = yaml.safe_load(res)
+        info("Loading versions from URL: " + versions_url)
+        VERSIONS = versions_yaml['versions']
+
+    except Exception:
+        try:
+            # try as a local file
+            with open(versions_url, 'r') as res:
+                versions_yaml = yaml.safe_load(res)
+                info("Loading versions from local file: " + versions_url)
+                VERSIONS = versions_yaml['versions']
+        except Exception:
+            info("Unable to load versions from path: " + versions_url)
+
+
 def provision(args, apic_file, no_random):
     config_file = args.config
     output_file = args.output
@@ -989,7 +900,7 @@ def provision(args, apic_file, no_random):
     if args.timeout:
         try:
             if int(args.timeout) >= 0:
-                timeout = args.timeout
+                timeout = int(args.timeout)
         except ValueError:
             # ignore that timeout value
             warn("Invalid timeout value ignored: '%s'" % timeout)
@@ -1015,6 +926,13 @@ def provision(args, apic_file, no_random):
             "debug_apic": args.debug,
         },
     }
+
+    # infra_vlan is not part of command line input, but we do
+    # pass it as a command line arg in unit tests to pass in
+    # configuration which would otherwise be discovered from
+    # the APIC
+    config["discovered"] = {"infra_vlan": getattr(args, "infra_vlan", None)}
+
     if args.username:
         config["aci_config"]["apic_login"]["username"] = args.username
 
@@ -1024,6 +942,12 @@ def provision(args, apic_file, no_random):
 
     # Create config
     user_config = config_user(config_file)
+
+    if user_config:
+        if 'versions_url' in user_config and 'path' in user_config['versions_url']:
+            versions_url = user_config['versions_url']['path']
+            get_versions(versions_url)
+
     deep_merge(config, user_config)
 
     flavor = DEFAULT_FLAVOR
@@ -1031,6 +955,7 @@ def provision(args, apic_file, no_random):
         flavor = args.flavor
     if flavor in FLAVORS:
         info("Using configuration flavor " + flavor)
+        deep_merge(config, {"flavor": flavor})
         if "config" in FLAVORS[flavor]:
             deep_merge(config, FLAVORS[flavor]["config"])
         if "default_version" in FLAVORS[flavor]:
@@ -1050,7 +975,8 @@ def provision(args, apic_file, no_random):
         deep_merge(config,
                    {"registry": VERSIONS[config["registry"]["version"]]})
 
-    deep_merge(config, config_discover(config, prov_apic))
+    # Discoverd state (e.g. infra-vlan) overrides the config file data
+    config = deep_merge(config_discover(config, prov_apic), config)
 
     # Validate APIC access
     if prov_apic is not None:
@@ -1092,13 +1018,27 @@ def provision(args, apic_file, no_random):
 
 def main(args=None, apic_file=None, no_random=False):
     # apic_file and no_random are used by the test functions
+    # len(sys.argv) == 1 when acc-provision is called w/o arguments
     if args is None:
-        args = parse_args()
+        args = parse_args(len(sys.argv) == 1)
+
+    if args.release:
+        try:
+            release_file_path = os.path.dirname(os.path.realpath(__file__)) + '/RELEASE-VERSION'
+            release = open(release_file_path, "r").read().rstrip()
+            print(release, file=sys.stderr)
+        except Exception:
+            info("Release info not present in this package")
+        return
 
     if args.list_flavors:
         info("Available configuration flavors:")
-        for flavor in FLAVORS:
-            info(flavor + ":\t" + FLAVORS[flavor]["desc"])
+        for flavor in sorted(FLAVORS, key=lambda x: int(FLAVORS[x]['order'])):
+            if not FLAVORS[flavor]['hidden']:
+                desc = FLAVORS[flavor]["desc"]
+                if FLAVORS[flavor]["status"]:
+                    desc = desc + " [" + FLAVORS[flavor]["status"] + "]"
+                info(flavor + ":\t" + desc)
         return
     if args.flavor is not None and args.flavor not in FLAVORS:
         err("Invalid configuration flavor: " + args.flavor)
